% MACHINE_DATA
% See also 
% 
%   Contents file for MACHINE_DATA and its subfolders.
%   
%   MACHINE_DATA
%   australian_synchrotron
%   dba        - Create dba lattice
%   esrf
%   FODO       - P1Dr=atdrift('Dr',0.1);
%   soleil     - Loads SOLEIL lattice
%   sp3v81f    - All the dipole and quadrupole lengths are effective lengths
%   spear2     - Example lattice definition file
%   spear2rad  - Example lattice definition file with CAVITY and CLASSICAL radiation
%   spear2resp - Example SPEAR2 lattice with orbit correctors and BPMS
%   spear3     - Load the SPEAR3 lattice structure
%   thomx
%    
%   This file was generated by updateContents.m on 06 Nov 2023 at 14:58:13.
